
module pwm_cont (
    input               reset_n,    // reset
    input               clk,        // clock
    input       [1:0]   addr,       // internal address
    input               sel,        // internal select
    input               write,      // internal write enable
    input       [31:0]  wdata,      // internal write data
    output      [31:0]  rdata,      // internal read data
    output              pwm_out     // pwm output
);

//-------------- internal signals --------------
    reg     [15:0]      div_reg;
    reg     [15:0]      per_reg;
    reg     [15:0]      duty_reg;

    reg     [15:0]      duty_reg_new ;
    
    wire                div_reg_set;
    wire                per_reg_set;
    
    reg                 cnt_ena;
    reg                 out_ena;
    
    reg     [15:0]      div_cnt;
    wire                div_cnt_over;
    reg     [15:0]      pwm_cnt;
    wire                pwm_cnt_over;    
    wire                pwm_out_clr;
    reg                 pwm_node;

	reg     [15:0]      rdata_mux ;
    
//-------------- logic description --------------

//-------------- register --------------
    always @(posedge clk or negedge reset_n) begin
        if(!reset_n) begin
            cnt_ena <= 1'b0;
            out_ena <= 1'b0;
            div_reg <= 0;
            per_reg <= 16'h01;
            duty_reg <= 0;
        end
        else if(sel && write) begin
            case(addr)
                2'b00 : begin
                    cnt_ena <= wdata[2];
                    out_ena <= wdata[4];
                end
                2'b01 : div_reg <= wdata[15:0];
                2'b10 : per_reg <= wdata[15:0];
                2'b11 : duty_reg <= wdata[15:0];
                default: ;
            endcase
        end
    end
 
    assign div_reg_set = sel && write && (addr == 2'b01);
    assign per_reg_set = sel && write && (addr == 2'b10);
    
//-------------- divisor counter --------------
    always @(posedge clk or negedge reset_n) begin
        if(!reset_n)
            div_cnt <= 0;
        else if(div_reg_set)
            div_cnt <= 0;
        else if(div_cnt_over)
            div_cnt <= 0;
        else
            div_cnt <= div_cnt + 1'b1;
    end
    
//-------------- pwm counter --------------
    assign div_cnt_over = (div_cnt == div_reg);
    assign pwm_cnt_over = (pwm_cnt == per_reg - 1'b1);
     
    always @(posedge clk or negedge reset_n) begin
        if(!reset_n)
            pwm_cnt <= 0;
        else if(!cnt_ena)
            pwm_cnt <= 0;
        else if(per_reg_set)
            pwm_cnt <= 0;
        else if(div_cnt_over) begin
            if(pwm_cnt_over)
                pwm_cnt <= 0;
            else
                pwm_cnt <= pwm_cnt + 1'b1;
        end
    end

//-------------- pwm output signal --------------
    assign pwm_out_clr = (pwm_cnt == duty_reg_new) || (duty_reg_new == 0) || (per_reg == 0) ;

    always @(posedge clk or negedge reset_n) begin
        if(!reset_n)
            duty_reg_new <= 0 ;
        else if(pwm_cnt==0)
            duty_reg_new <= duty_reg ;
    end

    always @(posedge clk or negedge reset_n) begin
        if(!reset_n)
            pwm_node <= 1'b0;
        else if(!cnt_ena)
            pwm_node <= 1'b0;
        else if(div_cnt_over) begin
            if(pwm_out_clr)
                pwm_node <= 1'b0;
            else if (pwm_cnt == 0)
                pwm_node <= 1'b1;
       end
    end
    
    assign pwm_out = (out_ena)? pwm_node : 1'b0;
    
//-------------- read data --------------
	assign rdata = {16'b0, rdata_mux};

    always @(posedge clk or negedge reset_n) begin
        if(!reset_n)
            rdata_mux <= 0;
        else begin
            case(addr)
                2'b00 : rdata_mux <= {11'b0, out_ena, 1'b0, cnt_ena, 2'b0};
                2'b01 : rdata_mux <= div_reg ;
                2'b10 : rdata_mux <= per_reg ;
                2'b11 : rdata_mux <= duty_reg ;
                default: rdata_mux <= 0;
            endcase
        end
    end

endmodule
