
module spi_cont (
    input               reset_n,    // reset
    input               clk,        // clock
    input       [1:0]   addr,       // internal address
    input               sel,        // internal select
    input               write,      // internal write enable
    input       [31:0]  wdata,      // internal write data
    output     [31:0]  rdata,      // internal read data
    output              irq,        // internal interrupt
    output              sclk,       // spi clock
    output              mosi,       // spi data output
    input               miso        // spi data input
);

//-------------- internal signals --------------
    reg                 inte;       // spi_ctr.interrupt enable
    reg                 cpol;       // spi_ctr.clock porarity
    reg                 cpha;       // spi_ctr.clock phase

    reg                 tx_emp;     // spi_str.tx empty
    reg                 tx_rdy;     // spi_str.tx ready

    reg         [7:0]   tx_reg;     // spi_dtr.txdata buffer
    reg         [7:0]   rx_reg;     // spi_dtr.rxdata buffer
    reg         [7:0]   br_reg;     // spi_br.bitrate
    reg         [7:0]   sf_reg;

    wire                tx_write;
    reg                 tx_write_d;
    wire                tx_end;
    wire                sf_write;
    
    reg         [7:0]   br_cnt;
    wire                br_cmp;
    
    reg         [3:0]   spi_cnt;

    reg                 sclk_d;
    reg                 miso_d;

	reg         [7:0]  rdata_mux ;
	
//-------------- logic description --------------

//-------------- write data --------------
    always @(posedge clk or negedge reset_n) begin
        if(!reset_n) begin
            inte <= 1'b0;
            cpol <= 1'b0;
            cpha <= 1'b0;
            tx_reg <= 0;
            br_reg <= 0;
        end
        else if(sel && write) begin
            case(addr)
                2'b00 : begin
                    inte <= wdata[7];
                    cpol <= wdata[2];
                    cpha <= wdata[1];
                end
                2'b10 : tx_reg <= wdata[7:0];
                2'b11 : br_reg <= wdata[7:0];
                default: ;
            endcase
        end
    end

//-------------- read data --------------
	assign rdata = {24'b0, rdata_mux} ;

    always @(posedge clk or negedge reset_n) begin
        if(!reset_n)
            rdata_mux <= 0;
        else begin
            case(addr)
                2'b00 : rdata_mux <= {inte, 4'b0000, cpol, cpha, 1'b0};
                2'b01 : rdata_mux <= {6'b000000, tx_emp, tx_rdy};
                2'b10 : rdata_mux <= rx_reg;
                2'b11 : rdata_mux <= br_reg;
                default: rdata_mux <= 0;
            endcase
        end
    end

    assign irq = (inte)? tx_emp : 1'b0;

//-------------- shift register write --------------
    assign tx_write = (sel & write & (addr == 2'b10));

    always @(posedge clk or negedge reset_n) begin
        if(!reset_n)
            tx_write_d <= 1'b0;
        else
            tx_write_d <= tx_write;
    end

    assign sf_write = (tx_rdy)? tx_write_d :
                      (!tx_emp)? tx_end : 1'b0;

//-------------- tx empty register --------------
    always @(posedge clk or negedge reset_n) begin
        if(!reset_n)
            tx_emp <= 1'b1;
        else if(tx_emp && tx_write)
            tx_emp <= 1'b0;
        else if(!tx_emp && sf_write)
            tx_emp <= 1'b1;
    end

//-------------- tx ready register --------------
    always @(posedge clk or negedge reset_n) begin
        if(!reset_n)
            tx_rdy <= 1'b1;
        else if(!tx_emp && sf_write)
            tx_rdy <= 1'b0;
        else if(tx_end)
            tx_rdy <= 1'b1;
    end

//-------------- bit rate divisor --------------
    always @(posedge clk or negedge reset_n) begin
        if(!reset_n)
            br_cnt <= 0;
        else if(tx_rdy)
            br_cnt <= 0;
        else if(br_cmp)
            br_cnt <= 0;
        else
            br_cnt <= br_cnt + 1'b1;
    end

    assign br_cmp = ((br_cnt == br_reg) & ~tx_rdy);

//-------------- spi timing counter --------------
    always @(posedge clk or negedge reset_n) begin
        if(!reset_n)
            spi_cnt <= 0;
        else if(sf_write)
            spi_cnt <= 0;
        else if(br_cmp)
            spi_cnt <= spi_cnt + 1'b1;
    end

//-------------- shift register --------------
    always @(posedge clk or negedge reset_n) begin
        if(!reset_n)
            sf_reg <= 0;
        else if(sf_write)
            sf_reg <= tx_reg;
        else if(br_cmp && spi_cnt[0])
            sf_reg <= {sf_reg[6:0], miso_d};
    end

    assign tx_end = ((spi_cnt == 15) & br_cmp);
    
    always @(posedge clk or negedge reset_n) begin
        if(!reset_n)
            rx_reg <= 0;
        else if(tx_end)
            rx_reg <= {sf_reg[6:0], miso_d};
    end
    
//-------------- spi signal --------------
    always @(posedge clk or negedge reset_n) begin
        if(!reset_n)
            sclk_d <= 1'b0;
        else if(tx_end && tx_emp)
            sclk_d <= 1'b0;
        else if(br_cmp)
            sclk_d <= ~sclk_d;
        else if(sf_write)
            sclk_d <= (cpha)? 1'b1 : 1'b0;
    end

    always @(posedge clk or negedge reset_n) begin
        if(!reset_n)
            miso_d <= 1'b0;
        else if(br_cmp && !spi_cnt[0])
            miso_d <= miso;
    end

//-------------- in/out dff --------------
    assign sclk = (cpol)? ~sclk_d : sclk_d;
    assign mosi = sf_reg[7];

endmodule
